# -*- coding: utf-8 -*-
from . import constants
import logging

__all__ = ["tts_langs"]

# Logger
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def tts_langs(tld="com"):
    """Languages Naver Text-to-Speech supports.

    Args:
        tld (string): Top-level domain for the Google Translate host
            to fetch languages from. i.e `https://translate.google.<tld>`.
            Default is ``com``.

    Returns:
        dict: A dictionnary of the type `{ '<lang>': '<name>'}`

        Where `<lang>` is an IETF language tag such as `en` or `pt-br`,
        and `<name>` is the full English name of the language, such as
        `English` or `Portuguese (Brazil)`.

    The dictionnary returned combines languages from two origins:

    - Languages fetched automatically from Google Translate
    - Languages that are undocumented variations that were observed to work and
      present different dialects or accents.

    """
    try:
        langs = dict()
        # log.debug("Fetching with '{}' tld".format(tld))
        # langs.update(_fetch_langs(tld))
        langs.update(_extra_langs())
        log.debug("langs: {}".format(langs))
        return langs
    except Exception as e:
        raise RuntimeError("Unable to get language list: {}".format(str(e)))


def _extra_langs():
    """Define extra languages.

    Returns:
        dict: A dictionnary of extra languages manually defined.
    """
    return constants.LANGUAGES



# -*- coding: utf-8 -*
from .core import PreProcessorRegex
from .core import PreProcessorSub
from .core import RegexBuilder
from .core import Tokenizer

from . import pre_processors
from . import tokenizer_cases

__all__ = [
    "RegexBuilder",
    "PreProcessorRegex",
    "PreProcessorSub",
    "Tokenizer",
    "pre_processors",
    "tokenizer_cases",
]



# -*- coding: utf-8 -*-
import re


class RegexBuilder:
    r"""Build regex using arguments passed into a pattern template.

    Builds a regex object for which the pattern is made from an argument
    passed into a template. If more than one argument is passed (iterable),
    each pattern is joined by "|" (regex alternation 'or') to create a
    single pattern.

    Args:
        pattern_args (iteratable): String element(s) to be each passed to
            ``pattern_func`` to create a regex pattern. Each element is
            ``re.escape``'d before being passed.
        pattern_func (callable): A 'template' function that should take a
            string and return a string. It should take an element of
            ``pattern_args`` and return a valid regex pattern group string.
        flags: ``re`` flag(s) to compile with the regex.

    Example:
        To create a simple regex that matches on the characters "a", "b",
        or "c", followed by a period::

            >>> rb = RegexBuilder('abc', lambda x: "{}\.".format(x))

        Looking at ``rb.regex`` we get the following compiled regex::

            >>> print(rb.regex)
            'a\.|b\.|c\.'

        The above is fairly simple, but this class can help in writing more
        complex repetitive regex, making them more readable and easier to
        create by using existing data structures.

    Example:
        To match the character following the words "lorem", "ipsum", "meili"
        or "koda"::

            >>> words = ['lorem', 'ipsum', 'meili', 'koda']
            >>> rb = RegexBuilder(words, lambda x: "(?<={}).".format(x))

        Looking at ``rb.regex`` we get the following compiled regex::

            >>> print(rb.regex)
            '(?<=lorem).|(?<=ipsum).|(?<=meili).|(?<=koda).'

    """

    def __init__(self, pattern_args, pattern_func, flags=0):
        """Create the RegexBuilder."""
        self.pattern_args = pattern_args
        self.pattern_func = pattern_func
        self.flags = flags

        # Compile
        self.regex = self._compile()

    def _compile(self):
        alts = []
        for arg in self.pattern_args:
            arg = re.escape(arg)
            alt = self.pattern_func(arg)
            alts.append(alt)

        pattern = "|".join(alts)
        return re.compile(pattern, self.flags)

    def __repr__(self):  # pragma: no cover
        """Print the regex."""
        return str(self.regex)


class PreProcessorRegex:
    r"""Regex-based substitution text pre-processor.

    Runs a series of regex substitutions (``re.sub``) from each ``regex`` of a
    :class:`gtts.tokenizer.core.RegexBuilder` with an extra ``repl``
    replacement parameter.

    Args:
        search_args (iteratable): String element(s) to be each passed to
            ``search_func`` to create a regex pattern. Each element is
            ``re.escape``'d before being passed.
        search_func (callable): A 'template' function that should take a
            string and return a string. It should take an element of
            ``search_args`` and return a valid regex search pattern string.
        repl (string): The common replacement passed to the ``sub`` method for
            each ``regex``. Can be a raw string (the case of a regex
            backreference, for example)
        flags: ``re`` flag(s) to compile with each `regex`.

    Example:
        Add "!" after the words "lorem" or "ipsum", while ignoring case::

            >>> import re
            >>> words = ['lorem', 'ipsum']
            >>> pp = PreProcessorRegex(words,
            ...                        lambda x: "({})".format(x), r'\\1!',
            ...                        re.IGNORECASE)

        In this case, the regex is a group and the replacement uses its
        backreference ``\\1`` (as a raw string). Looking at ``pp`` we get the
        following list of search/replacement pairs::

            >>> print(pp)
            (re.compile('(lorem)', re.IGNORECASE), repl='\1!'),
            (re.compile('(ipsum)', re.IGNORECASE), repl='\1!')

        It can then be run on any string of text::

            >>> pp.run("LOREM ipSuM")
            "LOREM! ipSuM!"

    See :mod:`gtts.tokenizer.pre_processors` for more examples.

    """

    def __init__(self, search_args, search_func, repl, flags=0):
        """Create the preprocessor."""
        self.repl = repl

        # Create regex list
        self.regexes = []
        for arg in search_args:
            rb = RegexBuilder([arg], search_func, flags)
            self.regexes.append(rb.regex)

    def run(self, text):
        """Run each regex substitution on ``text``.

        Args:
            text (string): the input text.

        Returns:
            string: text after all substitutions have been sequentially
            applied.

        """
        for regex in self.regexes:
            text = regex.sub(self.repl, text)
        return text

    def __repr__(self):  # pragma: no cover
        """Print the preprocessor."""
        subs_strs = []
        for r in self.regexes:
            subs_strs.append("({}, repl='{}')".format(r, self.repl))
        return ", ".join(subs_strs)


class PreProcessorSub:
    """Simple substitution text preprocessor.

    Performs string-for-string substitution from list a find/replace pairs.
    It abstracts :class:`gtts.tokenizer.core.PreProcessorRegex` with a default
    simple substitution regex.

    Args:
        sub_pairs (list): A list of tuples of the style
            ``(<search str>, <replace str>)``
        ignore_case (bool): Ignore case during search. Defaults to ``True``.

    Example:
        Replace all occurences of "Mac" to "PC" and "Firefox" to "Chrome"::

            >>> sub_pairs = [('Mac', 'PC'), ('Firefox', 'Chrome')]
            >>> pp = PreProcessorSub(sub_pairs)

        Looking at the ``pp``, we get the following list of
        search (regex)/replacement pairs::

            >>> print(pp)
            (re.compile('Mac', re.IGNORECASE), repl='PC'),
            (re.compile('Firefox', re.IGNORECASE), repl='Chrome')

        It can then be run on any string of text::

            >>> pp.run("I use firefox on my mac")
            "I use Chrome on my PC"

    See :mod:`gtts.tokenizer.pre_processors` for more examples.

    """

    def __init__(self, sub_pairs, ignore_case=True):
        """Create the preprocessor."""

        def search_func(x):
            return "{}".format(x)

        flags = re.I if ignore_case else 0

        # Create pre-processor list
        self.pre_processors = []
        for sub_pair in sub_pairs:
            pattern, repl = sub_pair
            pp = PreProcessorRegex([pattern], search_func, repl, flags)
            self.pre_processors.append(pp)

    def run(self, text):
        """Run each substitution on ``text``.

        Args:
            text (string): the input text.

        Returns:
            string: text after all substitutions have been sequentially
            applied.

        """
        for pp in self.pre_processors:
            text = pp.run(text)
        return text

    def __repr__(self):  # pragma: no cover
        """Print the preprocessor."""
        return ", ".join([str(pp) for pp in self.pre_processors])


class Tokenizer:
    r"""An extensible but simple generic rule-based tokenizer.

    A generic and simple string tokenizer that takes a list of functions
    (called `tokenizer cases`) returning ``regex`` objects and joins them by
    "|" (regex alternation 'or') to create a single regex to use with the
    standard ``regex.split()`` function.

    ``regex_funcs`` is a list of any function that can return a ``regex``
    (from ``re.compile()``) object, such as a
    :class:`gtts.tokenizer.core.RegexBuilder` instance (and its ``regex``
    attribute).

    See the :mod:`gtts.tokenizer.tokenizer_cases` module for examples.

    Args:
        regex_funcs (list): List of compiled ``regex`` objects. Each
            functions's pattern will be joined into a single pattern and
            compiled.
        flags: ``re`` flag(s) to compile with the final regex. Defaults to
            ``re.IGNORECASE``

    Note:
        When the ``regex`` objects obtained from ``regex_funcs`` are joined,
        their individual ``re`` flags are ignored in favour of ``flags``.

    Raises:
        TypeError: When an element of ``regex_funcs`` is not a function, or
            a function that does not return a compiled ``regex`` object.

    Warning:
        Joined ``regex`` patterns can easily interfere with one another in
        unexpected ways. It is recommanded that each tokenizer case operate
        on distinct or non-overlapping chracters/sets of characters
        (For example, a tokenizer case for the period (".") should also
        handle not matching/cutting on decimals, instead of making that
        a seperate tokenizer case).

    Example:
        A tokenizer with a two simple case (*Note: these are bad cases to
        tokenize on, this is simply a usage example*)::

            >>> import re, RegexBuilder
            >>>
            >>> def case1():
            ...     return re.compile("\,")
            >>>
            >>> def case2():
            ...     return RegexBuilder('abc', lambda x: "{}\.".format(x)).regex
            >>>
            >>> t = Tokenizer([case1, case2])

        Looking at ``case1().pattern``, we get::

            >>> print(case1().pattern)
            '\\,'

        Looking at ``case2().pattern``, we get::

            >>> print(case2().pattern)
            'a\\.|b\\.|c\\.'

        Finally, looking at ``t``, we get them combined::

            >>> print(t)
            're.compile('\\,|a\\.|b\\.|c\\.', re.IGNORECASE)
             from: [<function case1 at 0x10bbcdd08>, <function case2 at 0x10b5c5e18>]'

        It can then be run on any string of text::

            >>> t.run("Hello, my name is Linda a. Call me Lin, b. I'm your friend")
            ['Hello', ' my name is Linda ', ' Call me Lin', ' ', " I'm your friend"]

    """

    def __init__(self, regex_funcs, flags=re.IGNORECASE):
        """Create the tokenizer."""
        self.regex_funcs = regex_funcs
        self.flags = flags

        try:
            # Combine
            self.total_regex = self._combine_regex()
        except (TypeError, AttributeError) as e:  # pragma: no cover
            raise TypeError(
                "Tokenizer() expects a list of functions returning "
                "regular expression objects (i.e. re.compile). " + str(e)
            )

    def _combine_regex(self):
        alts = []
        for func in self.regex_funcs:
            alts.append(func())

        pattern = "|".join(alt.pattern for alt in alts)
        return re.compile(pattern, self.flags)

    def run(self, text):
        """Tokenize `text`.

        Args:
            text (string): the input text to tokenize.

        Returns:
            list: A list of strings (token) split according to the tokenizer cases.

        """
        return self.total_regex.split(text)

    def __repr__(self):  # pragma: no cover
        """Print the tokenizer."""
        return str(self.total_regex) + " from: " + str(self.regex_funcs)



# -*- coding: utf-8 -*-
import pytest
from navertts.lang import tts_langs, _extra_langs


"""Test language list downloading"""


def test_fetch_langs():
    """Fetch languages successfully"""
    # Downloaded Languages
    # Safe to assume 'en' (english) will always be there
    scraped_langs = {}  # _fetch_langs()
    # assert 'en' in scraped_langs

    # Scraping garbage
    # assert 'Detect language' not in scraped_langs
    # assert '—' not in scraped_langs

    # Add-in Languages
    all_langs = tts_langs()
    extra_langs = _extra_langs()
    assert len(all_langs) == len(scraped_langs) + len(extra_langs)


def test_fetch_langs_exception():
    """Raise RuntimeError on language fetch exception"""
    # with pytest.raises(RuntimeError):
    #     tts_langs(tld="invalid")
    pass


if __name__ == "__main__":
    pytest.main(["-x", __file__])



import warnings

TRANSLATE_ENDPOINT = "https://dict.naver.{tld}/api/nvoice"
TRANSLATE_PARAMS = (
    "?service=dictionary&speech_fmt=mp3&text={text}&speaker={speaker}&speed={speed}"
)

LANGUAGES = {
    "en": "English",
    "es": "Spanish",
    "ja": "Japanese",
    "ko": "Korean",
    "zh": "Chinese",
}

# https://apidocs.ncloud.com/en/ai-naver/clova_speech_synthesis/tts/
SPEAKERS = {
    "en": {"f": "clara", "m": "matt"},
    "es": {"f": "carmen", "m": "jose"},
    "ja": {"f": "nsayuri", "m": "shinji"},
    "ko": {"f": "kyuri", "m": "jinho"},
    "zh": {"f": "meimei", "m": "liangliang"},
}


def get_speaker(lang="ko", gender="f"):
    """Get the API name for the chosen speaker."""
    try:
        speakers = SPEAKERS[lang]
    except KeyError:
        raise ValueError(
            "No speaker for language {}. "
            "Available languages: {}".format(lang, list(SPEAKERS.keys()))
        )
    try:
        return speakers[gender]
    except KeyError:
        warnings.warn("Gender {} not available for language" " {}".format(gender, lang))
        return list(speakers.values())[0]


def translate_base(tld="com"):
    """Get the base URL."""
    return TRANSLATE_ENDPOINT.format(tld=tld)


def translate_endpoint(text, speaker="kyuri", speed=0, tld="com"):
    """Get the endpoint URL."""
    url = translate_base(tld=tld)
    return url + TRANSLATE_PARAMS.format(text=text, speaker=speaker, speed=speed)



# -*- coding: utf-8 -*-
from . import RegexBuilder
from . import symbols


def tone_marks():
    """Keep tone-modifying punctuation by matching following character.

    Assumes the `tone_marks` pre-processor was run for cases where there might
    not be any space after a tone-modifying punctuation mark.
    """
    return RegexBuilder(
        pattern_args=symbols.TONE_MARKS, pattern_func=lambda x: "(?<={}).".format(x)
    ).regex


def period_comma():
    """Period and comma case.

    Match if not preceded by ".<letter>" and only if followed by space.
    Won't cut in the middle/after dotted abbreviations; won't cut numbers.

    Note:
        Won't match if a dotted abbreviation ends a sentence.

    Note:
        Won't match the end of a sentence if not followed by a space.

    """
    return RegexBuilder(
        pattern_args=symbols.PERIOD_COMMA,
        pattern_func=lambda x: r"(?<!\.[a-z]){} ".format(x),
    ).regex


def colon():
    """Colon case.

    Match a colon ":" only if not preceeded by a digit.
    Mainly to prevent a cut in the middle of time notations e.g. 10:01

    """
    return RegexBuilder(
        pattern_args=symbols.COLON, pattern_func=lambda x: r"(?<!\d){}".format(x)
    ).regex


def other_punctuation():
    """Match other punctuation.

    Match other punctuation to split on; punctuation that naturally
    inserts a break in speech.

    """
    punc = "".join(
        set(symbols.ALL_PUNC)
        - set(symbols.TONE_MARKS)
        - set(symbols.PERIOD_COMMA)
        - set(symbols.COLON)
    )
    return RegexBuilder(pattern_args=punc, pattern_func=lambda x: "{}".format(x)).regex


def legacy_all_punctuation():  # pragma: no cover b/c tested but Coveralls: ¯\_(ツ)_/¯
    """Match all punctuation.

    Use as only tokenizer case to mimic gTTS 1.x tokenization.
    """
    punc = symbols.ALL_PUNC
    return RegexBuilder(pattern_args=punc, pattern_func=lambda x: "{}".format(x)).regex



# -*- coding: utf-8 -*-
import pytest
from navertts.utils import _minimize, _len, _clean_tokens
from navertts.constants import translate_endpoint

delim = " "
Lmax = 10


def test_ascii():
    _in = "Bacon ipsum dolor sit amet"
    _out = ["Bacon", "ipsum", "dolor sit", "amet"]
    assert _minimize(_in, delim, Lmax) == _out


def test_ascii_no_delim():
    _in = "Baconipsumdolorsitametflankcornedbee"
    _out = ["Baconipsum", "dolorsitam", "etflankcor", "nedbee"]
    assert _minimize(_in, delim, Lmax) == _out


def test_unicode():
    _in = "这是一个三岁的小孩在讲述他从一系列照片里看到的东西。"
    _out = ["这是一个三岁的小孩在", "讲述他从一系列照片里", "看到的东西。"]
    assert _minimize(_in, delim, Lmax) == _out


def test_startwith_delim():
    _in = delim + "test"
    _out = ["test"]
    assert _minimize(_in, delim, Lmax) == _out


def test_len_ascii():
    text = "Bacon ipsum dolor sit amet flank corned beef."
    assert _len(text) == 45


def test_len_unicode():
    text = "但在一个重要的任务上"
    assert _len(text) == 10


def test_only_space_and_punc():
    _in = [",(:)?", "\t    ", "\n"]
    _out = []
    assert _clean_tokens(_in) == _out


def test_strip():
    _in = [" Bacon  ", "& ", "ipsum\r", "."]
    _out = ["Bacon", "&", "ipsum"]
    assert _clean_tokens(_in) == _out


def test_translate_url():
    _in = {"tld": "qwerty", "text": "uiop", "speed": 5, "speaker": "foobar"}
    _out = (
        "https://dict.naver.qwerty/api/nvoice"
        "?service=dictionary&speech_fmt=mp3&text=uiop&speaker=foobar&speed=5"
    )
    assert translate_endpoint(**_in) == _out


if __name__ == "__main__":
    pytest.main(["-x", __file__])



# -*- coding: utf-8 -*-
from . import __version__
from . import NaverTTS
from . import NaverTTSError
from .lang import tts_langs
import click
import logging
import logging.config

# Click settings
CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}

# Logger settings
LOGGER_SETTINGS = {
    "version": 1,
    "formatters": {"default": {"format": "%(name)s - %(levelname)s - %(message)s"}},
    "handlers": {"console": {"class": "logging.StreamHandler", "formatter": "default"}},
    "loggers": {"navertts": {"handlers": ["console"], "level": "WARNING"}},
}

# Logger
logging.config.dictConfig(LOGGER_SETTINGS)
log = logging.getLogger("navertts")


def sys_encoding():
    """Charset to use for --file <path>|- (stdin)."""
    return "utf8"


def validate_text(ctx, param, text):
    """Validate <text> argument.

    Ensures <text> (arg) and <file> (opt) are mutually exclusive
    """
    if not text and "file" not in ctx.params:
        # No <text> and no <file>
        return "-"
    elif text and "file" in ctx.params:
        # Both <text> and <file>
        raise click.BadParameter("<text> and -f/--file <file> can't be used together")
    elif text and isinstance(text, tuple):
        return " ".join(text)
    else:
        return text


def validate_lang(ctx, param, lang):
    """Validate <lang> option.

    Ensures <lang> is a supported language unless the <nocheck> flag is set
    Uses <tld> to fetch languages from other domains
    """
    if ctx.params["nocheck"]:
        return lang

    try:
        tld = ctx.params["tld"]
        if lang not in tts_langs(tld):
            raise click.UsageError(
                "'%s' not in list of supported languages.\n"
                "Use --all to list languages or "
                "add --nocheck to disable language check." % lang
            )
        else:
            # The language is valid.
            # No need to let NaverTTS re-validate.
            ctx.params["nocheck"] = True
    except RuntimeError as e:
        # Only case where the <nocheck> flag can be False
        # Non-fatal. gTTS will try to re-validate.
        log.debug(str(e), exc_info=True)

    return lang


def validate_speed(ctx, param, speed):
    """Validate <speed> option.

    Ensures <speed> is a supported language unless the <nocheck> flag is set
    Uses <tld> to fetch languages from other domains
    """
    if speed in ["slow", "normal", "fast"]:
        return speed

    try:
        return int(speed)
    except TypeError:
        raise click.UsageError(
            "'%s' not in list of supported speeds.\n"
            "Choose from ['slow', 'normal', 'fast'] or "
            "an integer between -5 (fast) and 5 (slow)." % speed
        )


def print_languages(ctx, param, value):
    """Print all languages.

    Prints formatted sorted list of supported languages and exits
    """
    if not value or ctx.resilient_parsing:
        return

    try:
        tld = ctx.params["tld"]
    except KeyError:
        # Either --tld was used after --all or not at all
        # Default to the 'com' tld
        tld = "com"

    try:
        langs = tts_langs(tld)
        langs_str_list = sorted("{}: {}".format(k, langs[k]) for k in langs)
        click.echo("  " + "\n  ".join(langs_str_list))
    except RuntimeError as e:  # pragma: no cover
        log.debug(str(e), exc_info=True)
        raise click.ClickException("Couldn't fetch language list.")
    ctx.exit()


def set_debug(ctx, param, debug):
    """Set logger level to DEBUG."""
    if debug:
        log.setLevel(logging.DEBUG)
    return


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "text", metavar="<text>", nargs=-1, required=False, callback=validate_text
)
@click.option(
    "-f",
    "--file",
    metavar="<file>",
    # For py2.7/unicode. If encoding not None Click uses io.open
    type=click.File(encoding=sys_encoding()),
    help="Read from <file> instead of <text>.",
)
@click.option(
    "-o",
    "--output",
    metavar="<file>",
    type=click.File(mode="wb"),
    help="Write to <file> instead of stdout.",
)
@click.option(
    "-s",
    "--speed",
    metavar="<speed>",
    default="normal",
    show_default=True,
    callback=validate_speed,
    help="Reading speed. Choose from 'slow', 'normal', 'fast' "
    "or an integer between -5 (fast) and 5 (slow).",
)
@click.option(
    "-l",
    "--lang",
    metavar="<lang>",
    default="ko",
    show_default=True,
    callback=validate_lang,
    help="IETF language tag. Language to speak in. List documented tags with --all.",
)
@click.option(
    "-t",
    "--tld",
    metavar="<tld>",
    default="com",
    show_default=True,
    is_eager=True,  # Prioritize <tld> to ensure it gets set before <lang>
    help="Top-level domain for the Google host, i.e https://translate.google.<tld>",
)
@click.option(
    "--nocheck",
    default=False,
    is_flag=True,
    is_eager=True,  # Prioritize <nocheck> to ensure it gets set before <lang>
    help="Disable strict IETF language tag checking. Allow undocumented tags.",
)
@click.option(
    "--all",
    default=False,
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=print_languages,
    help="Print all documented available IETF language tags and exit. "
    "Use --tld beforehand to use an alternate domain",
)
@click.option(
    "--debug",
    default=False,
    is_flag=True,
    is_eager=True,  # Prioritize <debug> to see debug logs of callbacks
    expose_value=False,
    callback=set_debug,
    help="Show debug information.",
)
@click.version_option(version=__version__)
def tts_cli(text, file, output, speed, tld, lang, nocheck):
    """Read <text> to mp3 format using NAVER Papago's Text-to-Speech API.

    (set <text> or --file <file> to - for standard input)
    """
    # stdin for <text>
    if text == "-":
        text = click.get_text_stream("stdin").read()

    # stdout (when no <output>)
    if not output:
        output = click.get_binary_stream("stdout")

    # <file> input (stdin on '-' is handled by click.File)
    if file:
        try:
            text = file.read()
        except UnicodeDecodeError as e:  # pragma: no cover
            log.debug(str(e), exc_info=True)
            raise click.FileError(
                file.name, "<file> must be encoded using '%s'." % sys_encoding()
            )

    # TTS
    try:
        tts = NaverTTS(
            text=text, lang=lang, speed=speed, tld=tld, lang_check=not nocheck
        )
        tts.write_to_fp(output)
    except (ValueError, AssertionError) as e:
        raise click.UsageError(str(e))
    except NaverTTSError as e:
        raise click.ClickException(str(e))



# -*- coding: utf-8 -*-
from .tokenizer.symbols import ALL_PUNC as punc
import re
from string import whitespace

_ALL_PUNC_OR_SPACE = re.compile("^[{}]*$".format(re.escape(punc + whitespace)))
"""Regex that matches if an entire line is only comprised
of whitespace and punctuation

"""


def _minimize(the_string, delim, max_size):
    """Recursively split a string.

    Splits a string in the largest chunks
    possible from the highest position of a delimiter all the way
    to a maximum size

    Args:
        the_string (string): The string to split.
        delim (string): The delimiter to split on.
        max_size (int): The maximum size of a chunk.

    Returns:
        list: the minimized string in tokens

    Every chunk size will be at minimum ``the_string[0:idx]`` where ``idx``
    is the highest index of ``delim`` found in ``the_string``; and at maximum
    ``the_string[0:max_size]`` if no ``delim`` was found in ``the_string``.
    In the latter case, the split will occur at ``the_string[max_size]``
    which can be any character. The function runs itself again on the rest of
    ``the_string`` (``the_string[idx:]``) until no chunk is larger than
    ``max_size``.

    """
    # Remove `delim` from start of `the_string`
    # i.e. prevent a recursive infinite loop on `the_string[0:0]`
    # if `the_string` starts with `delim` and is larger than `max_size`
    if the_string.startswith(delim):
        the_string = the_string[_len(delim) :]

    if _len(the_string) > max_size:
        try:
            # Find the highest index of `delim` in `the_string[0:max_size]`
            # i.e. `the_string` will be cut in half on `delim` index
            idx = the_string.rindex(delim, 0, max_size)
        except ValueError:
            # `delim` not found in `the_string`, index becomes `max_size`
            # i.e. `the_string` will be cut in half arbitrarily on `max_size`
            idx = max_size
        # Call itself again for `the_string[idx:]`
        return [the_string[:idx]] + _minimize(the_string[idx:], delim, max_size)
    else:
        return [the_string]


def _len(text):
    """Compute the length of a string.

    Same as ``len(text)`` for a string but that decodes
    ``text`` first in Python 2.x

    Args:
        text (string): String to get the size of.

    Returns:
        int: The size of the string.
    """
    try:
        # Python 2
        return len(unicode(text))
    except NameError:  # pragma: no cover
        # Python 3
        return len(text)


def _clean_tokens(tokens):
    """Clean a list of strings.

    Args:
        tokens (list): A list of strings (tokens) to clean.

    Returns:
        list: Stripped strings ``tokens`` without the original elements
            that only consisted of whitespace and/or punctuation characters.

    """
    return [t.strip() for t in tokens if not _ALL_PUNC_OR_SPACE.match(t)]


def _quote_all(input_string, *args, **kwargs):  # pylint:disable=unused-argument
    """NAVER Translate needs every character quoted."""
    return "".join("%%%x" % ord(char) for char in input_string)


def _log(log_fun, input_string, *args):
    """Log message, sanitizing string if necessary."""
    try:
        return log_fun(input_string, *args)
    except UnicodeEncodeError:
        input_string = input_string.encode("ascii", "xmlcharrefreplace").decode()
        return log_fun(input_string, *args)



from setuptools import setup, find_packages
import os

version_py = os.path.join(os.path.dirname(__file__), "navertts", "version.py")
version = open(version_py).read().strip().split("=")[-1].replace('"', "").strip()

setup(
    name="NaverTTS",
    packages=find_packages(),
    python_requires=">= 3.5",
    include_package_data=True,
    install_requires=[
        "six",
        "beautifulsoup4",
        "click",
        "requests",
    ],
    extras_require={
        "test": [
            "pytest >= 4.6",
            "pytest-cov",
            "flake8",
            "testfixtures",
            "mock",
            "coveralls",
        ]
    },
    entry_points={"console_scripts": ["navertts-cli=navertts.cli:tts_cli"]},
    description="NaverTTS (NAVER Text-to-Speech), a Python library and CLI tool to "
    "interface with NAVER Papago text-to-speech API",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    author="Scott Gigante",
    author_email="scottgigante@gmail.com",
    url="https://github.com/scottgigante/NaverTTS",
    version=version,
    test_suite="navertts.tests",
    keywords=[
        "navertts",
        "text to speech",
        "NAVER Papago",
        "TTS",
    ],
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
        "Topic :: Multimedia :: Sound/Audio :: Speech",
    ],
    license="MIT",
)



# -*- coding: utf-8 -*-
import os
import pytest
from mock import Mock

from navertts.tts import NaverTTS, NaverTTSError
from navertts.lang import _extra_langs

# Testing all languages takes some time.
# Set TEST_LANGS envvar to choose languages to test.
#  * 'fetch': Languages fetched from the Web
#  * 'extra': Languagee set in Languages.EXTRA_LANGS
#  * 'all': All of the above
#  * <csv>: Languages tags list to test
# Unset TEST_LANGS to test everything ('all')
# See: langs_dict()


"""Construct a dict of suites of languages to test.
{ '<suite name>' : <list or dict of language tags> }

ex.: { 'fetch' : {'en': 'English', 'fr': 'French'},
       'extra' : {'en': 'English', 'fr': 'French'} }
ex.: { 'environ' : ['en', 'fr'] }
"""
env = os.environ.get("TEST_LANGS")
if not env or env == "all":
    langs = {}  # _fetch_langs()
    langs.update(_extra_langs())
elif env == "fetch":
    langs = {}  # _fetch_langs()
elif env == "extra":
    langs = _extra_langs()
else:
    env_langs = {lang: lang for lang in env.split(",") if lang}
    langs = env_langs


@pytest.mark.parametrize("lang", langs.keys(), ids=list(langs.values()))
def test_TTS(tmp_path, lang):
    """Test all supported languages and file save."""
    text = "This is a test"
    """Create output .mp3 file successfully"""
    for speed in ("normal", "slow", "fast"):
        filename = tmp_path / "test_{}_.mp3".format(lang)
        # Create NaverTTS and save
        tts = NaverTTS(text=text, lang=lang, speed=speed)
        tts.save(filename)

        # Check if files created is > 2k
        assert filename.stat().st_size > 2000


def test_unsupported_language_check():
    """Raise ValueError on unsupported language (with language check)."""
    lang = "xx"
    text = "Lorem ipsum"
    check = True
    with pytest.raises(ValueError):
        NaverTTS(text=text, lang=lang, lang_check=check)


def test_empty_string():
    """Raise AssertionError on empty string."""
    text = ""
    with pytest.raises(AssertionError):
        NaverTTS(text=text)


def test_no_text_parts(tmp_path):
    """Raises AssertionError on no content to send to API (no text_parts)."""
    text = "                                                                                                          ..,\n"
    with pytest.raises(AssertionError):
        filename = tmp_path / "no_content.txt"
        tts = NaverTTS(text=text)
        tts.save(filename)


# Test write_to_fp()/save() cases not covered elsewhere in this file


def test_bad_fp_type():
    """Raise TypeError if fp is not a file-like object (no .write())."""
    # Create gTTS and save
    tts = NaverTTS(text="test")
    with pytest.raises(TypeError):
        tts.write_to_fp(5)


def test_save(tmp_path):
    """Save .mp3 file successfully."""
    filename = tmp_path / "save.mp3"
    # Create NaverTTS and save
    tts = NaverTTS(text="test")
    tts.save(filename)

    # Check if file created is > 2k
    assert filename.stat().st_size > 2000


def test_msg():
    """Test NaverTTSError internal exception handling.

    Set exception message successfully.
    """
    error1 = NaverTTSError("test")
    assert "test" == error1.msg

    error2 = NaverTTSError()
    assert error2.msg is None


def test_infer_msg():
    """Infer message sucessfully based on context."""
    # Without response:

    # Bad TLD
    ttsTLD = Mock(tld="invalid")
    errorTLD = NaverTTSError(tts=ttsTLD)
    assert (
        errorTLD.msg
        == "Failed to connect. Probable cause: Host 'https://dict.naver.invalid/api/nvoice' is not reachable"
    )

    # With response:

    # 403
    tts403 = Mock()
    response403 = Mock(status_code=403, reason="aaa")
    error403 = NaverTTSError(tts=tts403, response=response403)
    assert (
        error403.msg
        == "403 (aaa) from TTS API. Probable cause: Bad token or upstream API changes"
    )

    # 404 (and not lang_check)
    tts404 = Mock(lang="xx", lang_check=False)
    response404 = Mock(status_code=404, reason="bbb")
    error404 = NaverTTSError(tts=tts404, response=response404)
    assert (
        error404.msg
        == "404 (bbb) from TTS API. Probable cause: Unsupported language 'xx'"
    )

    # >= 500
    tts500 = Mock()
    response500 = Mock(status_code=500, reason="ccc")
    error500 = NaverTTSError(tts=tts500, response=response500)
    assert (
        error500.msg
        == "500 (ccc) from TTS API. Probable cause: Uptream API error. Try again later."
    )

    # Unknown (ex. 100)
    tts100 = Mock()
    response100 = Mock(status_code=100, reason="ddd")
    error100 = NaverTTSError(tts=tts100, response=response100)
    assert error100.msg == "100 (ddd) from TTS API. Probable cause: Unknown"


def test_WebRequest(tmp_path):
    """Test Web Requests."""
    text = "Lorem ipsum"

    """Raise NaverTTSError on unsupported language (without language check)"""
    lang = "xx"
    check = False

    with pytest.raises(ValueError):
        filename = tmp_path / "xx.txt"
        # Create NaverTTS
        tts = NaverTTS(text=text, lang=lang, lang_check=check)
        tts.save(filename)


if __name__ == "__main__":
    pytest.main(["-x", __file__])



# -*- coding: utf-8 -*-
from . import constants
from . import tokenizer
from . import utils
from .lang import tts_langs

import logging
import os
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import urllib

__all__ = ["NaverTTS", "NaverTTSError"]

# Logger
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class Speed:
    """Read Speed."""

    SLOW = 5
    NORMAL = 0
    FAST = -5


class NaverTTS:
    """NaverTTS -- NAVER Text-to-Speech.

    An interface to NAVER Papago's Text-to-Speech API.

    Args:
        text (string): The text to be read.
        tld (string, optional): Top-level domain. Defaults to 'com'.
        lang (string, optional): The language (IETF language tag) to
            read the text in. Defaults to 'ko'.
        speed (str or int, optional): Choice of {'slow' (5), 'normal' (0), 'fast' (-5)}.
            Defaults to 'normal'.
        lang_check (bool, optional): Strictly enforce an existing ``lang``,
            to catch a language error early. If set to ``True``,
            a ``ValueError`` is raised if ``lang`` doesn't exist.
            Default is ``True``.
        pre_processor_funcs (list): A list of zero or more functions that are
            called to transform (pre-process) text before tokenizing. Those
            functions must take a string and return a string. Defaults to::

                [
                    tokenizer.pre_processors.tone_marks,
                    tokenizer.pre_processors.end_of_line,
                    tokenizer.pre_processors.abbreviations,
                    tokenizer.pre_processors.word_sub
                ]

        tokenizer_func (callable): A function that takes in a string and
            returns a list of string (tokens). Defaults to::

                tokenizer.Tokenizer([
                    tokenizer.tokenizer_cases.tone_marks,
                    tokenizer.tokenizer_cases.period_comma,
                    tokenizer.tokenizer_cases.colon,
                    tokenizer.tokenizer_cases.other_punctuation
                ]).run

    See Also:
        :doc:`Pre-processing and tokenizing <tokenizer>`

    Raises:
        AssertionError: When ``text`` is ``None`` or empty; when there's nothing
            left to speak after pre-precessing, tokenizing and cleaning.
        ValueError: When ``lang_check`` is ``True`` and ``lang`` is not supported.
        RuntimeError: When ``lang_check`` is ``True`` but there's an error loading
            the languages dictionnary.

    """

    NAVER_TTS_MAX_CHARS = 100  # Max characters the NAVER TTS API takes at a time
    NAVER_TTS_HEADERS = {
        "Referer": "http://papago.naver.com/",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/47.0.2526.106 Safari/537.36",
    }

    def __init__(
        self,
        text,
        lang="ko",
        tld="com",
        speed="normal",
        gender="f",
        lang_check=True,
        pre_processor_funcs=[
            tokenizer.pre_processors.tone_marks,
            tokenizer.pre_processors.end_of_line_hyphen,
            tokenizer.pre_processors.newline,
            tokenizer.pre_processors.abbreviations,
            tokenizer.pre_processors.word_sub,
        ],
        tokenizer_func=tokenizer.Tokenizer(
            [
                tokenizer.tokenizer_cases.tone_marks,
                tokenizer.tokenizer_cases.period_comma,
                tokenizer.tokenizer_cases.colon,
                tokenizer.tokenizer_cases.other_punctuation,
            ]
        ).run,
    ):
        """Create the TTS class."""
        # Debug
        for k, v in locals().items():
            if k == "self":
                continue
            utils._log(log.debug, "%s: %s", k, v)

        # Text
        assert text, "No text to speak"
        self.text = text

        # Translate URL top-level domain
        self.tld = tld

        # Language
        if lang_check:
            try:
                langs = tts_langs(self.tld)
                if lang.lower() not in langs:
                    raise ValueError("Language not supported: %s" % lang)
            except RuntimeError as e:
                utils._log(log.debug, str(e), exc_info=True)
                utils._log(log.warning, str(e))

        self.lang_check = lang_check
        self.lang = lang.lower()
        self.speaker = constants.get_speaker(self.lang, gender)

        # Read speed
        if speed == "slow":
            self.speed = Speed.SLOW
        elif speed == "normal":
            self.speed = Speed.NORMAL
        elif speed == "fast":
            self.speed = Speed.FAST
        elif isinstance(speed, int) and -5 <= speed and speed <= 5:
            self.speed = speed
        else:
            raise ValueError(
                "Expected `speed` in 'slow', 'normal', "
                "'fast' or an integer between -5 and 5."
                " Got {}".format(speed)
            )

        # Pre-processors and tokenizer
        self.pre_processor_funcs = pre_processor_funcs
        self.tokenizer_func = tokenizer_func

    def _tokenize(self, text):
        # Pre-clean
        text = text.strip()

        # Apply pre-processors
        for pp in self.pre_processor_funcs:
            utils._log(log.debug, "pre-processing: %s", pp)
            text = pp(text)

        if utils._len(text) <= self.NAVER_TTS_MAX_CHARS:
            return utils._clean_tokens([text])

        # Tokenize
        utils._log(log.debug, "tokenizing: %s", self.tokenizer_func)
        tokens = self.tokenizer_func(text)

        # Clean
        tokens = utils._clean_tokens(tokens)

        # Minimize
        min_tokens = []
        for t in tokens:
            min_tokens += utils._minimize(t, " ", self.NAVER_TTS_MAX_CHARS)
        return min_tokens

    def write_to_fp(self, fp):
        """Do the TTS API request and write bytes to a file-like object.

        Args:
            fp (file object): Any file-like object to write the ``mp3`` to.

        Raises:
            :class:`gTTSError`: When there's an error with the API request.
            TypeError: When ``fp`` is not a file-like object that takes bytes.

        """
        # When disabling ssl verify in requests (for proxies and firewalls),
        # urllib3 prints an insecure warning on stdout. We disable that.
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

        text_parts = self._tokenize(self.text)
        utils._log(log.debug, "text_parts: %i", len(text_parts))
        assert text_parts, "No text to send to TTS API"

        for idx, part in enumerate(text_parts):

            endpoint_url = constants.translate_endpoint(
                text=part, speaker=self.speaker, speed=self.speed, tld=self.tld
            )
            try:
                # Request
                r = requests.get(
                    url=endpoint_url,
                    headers=self.NAVER_TTS_HEADERS,
                    proxies=urllib.request.getproxies(),
                    verify=False,
                )

                utils._log(log.debug, "headers-%i: %s", idx, r.request.headers)
                utils._log(log.debug, "url-%i: %s", idx, r.request.url)
                utils._log(log.debug, "status-%i: %s", idx, r.status_code)

                r.raise_for_status()
            except requests.exceptions.HTTPError as e:  # pragma: no cover
                # Request successful, bad response
                utils._log(log.debug, str(e))
                raise NaverTTSError(tts=self, response=r)
            except requests.exceptions.RequestException as e:  # pragma: no cover
                # Request failed
                utils._log(log.debug, str(e))
                raise NaverTTSError(tts=self)

            try:
                for chunk in r.iter_content(chunk_size=1024):
                    fp.write(chunk)
                utils._log(log.debug, "part-%i written to %s", idx, fp)
            except (AttributeError, TypeError) as e:
                raise TypeError(
                    "'fp' is not a file-like object or it does not take bytes: %s"
                    % str(e)
                )

    def save(self, savefile):
        """Do the TTS API request and write result to file.

        Args:
            savefile (string): The path and file name to save the ``mp3`` to.

        Raises:
            :class:`NaverTTSError`: When there's an error with the API request.

        """
        savefile = str(savefile)
        try:
            with open(savefile, "wb") as f:
                self.write_to_fp(f)
                utils._log(log.debug, "Saved to %s", savefile)
        except NaverTTSError:
            os.remove(savefile)
            raise


class NaverTTSError(Exception):
    """Exception that uses context to present a meaningful error message."""

    def __init__(self, msg=None, **kwargs):
        """Create a TTS exception."""
        self.tts = kwargs.pop("tts", None)
        self.rsp = kwargs.pop("response", None)
        if msg:
            self.msg = msg
        elif self.tts is not None:
            self.msg = self.infer_msg(self.tts, self.rsp)
        else:
            self.msg = None
        super(NaverTTSError, self).__init__(self.msg)

    def infer_msg(self, tts, rsp=None):
        """Attempt to guess what went wrong.

        Uses known information (e.g. http response) and observed behaviour.

        """
        cause = "Unknown"

        if rsp is None:
            premise = "Failed to connect"

            if tts.tld != "com":
                host = constants.translate_base(tld=tts.tld)
                cause = "Host '{}' is not reachable".format(host)

        else:
            # rsp should be <requests.Response>
            # http://docs.python-requests.org/en/master/api/
            status = rsp.status_code
            reason = rsp.reason

            premise = "{:d} ({}) from TTS API".format(status, reason)

            if status == 403:
                cause = "Bad token or upstream API changes"
            elif status == 404 and not tts.lang_check:
                cause = "Unsupported language '%s'" % self.tts.lang
            elif status >= 500:
                cause = "Uptream API error. Try again later."

        return "{}. Probable cause: {}".format(premise, cause)
